﻿Imports Microsoft.VisualBasic
Imports System
Imports System.Collections.Specialized
Imports System.Collections.Generic
Imports System.Configuration
Imports System.Data.Caching

Imports NorthwindCacheDemo.DTO
Imports NorthwindCacheDemo.DAL

Namespace NorthwindCacheDemo.CL.Velocity
	''' <summary>
	''' Provides a cache layer for the Northwind product catalog.
	''' VelocityCatalogCache acts as a read-through layer in which
	''' operations that result in a cache miss will subsequently 
	''' populate the cache by generating calls to the underlying
	''' data store. Methods in this class first attempt to retrieve
	''' products and categories from the catalog and reports named
	''' caches. If these methods result in a cache miss, they will
	''' call into the data layer to populate the cache and provide
	''' the caller with the result. These methods do not provide
	''' concurrency support. As this data is primarily read-only,
	''' the impact will not be too dramatic; however, for a more
	''' robust approach to concurrency, these methods could take
	''' locks before going to the data tier if a cache miss occurs.
	''' For more information on concurrency, please see the
	''' documentation provided with the Velocity installation.
	''' </summary>
	Public Class VelocityCatalogCache
		Implements ICatalogCache
		Private Const CATEGORY_KEY_FORMAT As String = "Category-{0}"
		Private Const PRODUCT_KEY_FORMAT As String = "Product-{0}"
		Private Const PRODUCT_LIST_KEY_FORMAT As String = "ProductList-{0}"
		Private Const POPULAR_LIST_KEY_FORMAT As String = "PopularProductList-{0}"
		Private Const ALSO_PURCHASED_KEY_FORMAT As String = "AlsoPurchasedProductList-{0}"
		Private Const CATEGORY_LIST_KEY As String = "CategoryList"
		Private Const POPULAR_LIST_KEY As String = "PopularProductList"

		Private _catalogCache As Cache

		Private _reportCache As Cache

		Public Sub New()
			Dim config As NameValueCollection = CType(ConfigurationManager.GetSection("northwindCacheDemo"), NameValueCollection)

			' Create end point for local caching
			Dim endPoint As New ServerEndPoint(config("VelocityHostName"), Convert.ToInt32(config("VelocityPort")), config("VelocityCacheHostName"))

			' Create catalog cache client with local caching
			Dim factory As New CacheFactory(New ServerEndPoint() { endPoint }, True, True)
			_catalogCache = factory.GetCache("catalog")

			' Create reports cache client without local caching
			factory = New CacheFactory()
			_reportCache = factory.GetCache("reports")
		End Sub

		Public Function GetCategories() As IList(Of Category) Implements ICatalogCache.GetCategories
			Dim categoryIDs As IList(Of Integer) = CType(_catalogCache.Get(CATEGORY_LIST_KEY), IList(Of Integer))
			Dim categories As IList(Of Category) = Nothing

			If categoryIDs IsNot Nothing Then
				categories = New List(Of Category)(categoryIDs.Count)
				For Each id As Integer In categoryIDs
					categories.Add(CType(_catalogCache.Get(String.Format(CATEGORY_KEY_FORMAT, id)), Category))
				Next id
			Else
				categories = DataLayer.CatalogProvider.GetCategories()

				If categories IsNot Nothing AndAlso categories.Count > 0 Then
					categoryIDs = New List(Of Integer)(categories.Count)
					For Each c As Category In categories
						categoryIDs.Add(c.CategoryID)
						_catalogCache.Put(String.Format(CATEGORY_KEY_FORMAT, c.CategoryID), c)
					Next c

					_catalogCache.Put(CATEGORY_LIST_KEY, categoryIDs)
				End If
			End If

			Return categories
		End Function

		Public Function GetCategory(ByVal id As Integer) As Category Implements ICatalogCache.GetCategory
			Dim key As String = String.Format(CATEGORY_KEY_FORMAT, id)
			Dim c As Category = CType(_catalogCache.Get(key), Category)

			If c Is Nothing Then
				c = DataLayer.CatalogProvider.GetCategory(id)

				If c IsNot Nothing Then
					_catalogCache.Put(key, c)
				End If
			End If

			Return c
		End Function

		Public Function GetProduct(ByVal id As Integer) As Product Implements ICatalogCache.GetProduct
			Dim key As String = String.Format(PRODUCT_KEY_FORMAT, id)
			Dim p As Product = CType(_catalogCache.Get(key), Product)

			If p Is Nothing Then
				p = DataLayer.CatalogProvider.GetProduct(id)

				If p IsNot Nothing Then
					_catalogCache.Put(key, p)
				End If
			End If

			Return p
		End Function

		Public Function GetProductsByCategory(ByVal id As Integer) As IList(Of Product) Implements ICatalogCache.GetProductsByCategory
			Dim listKey As String = String.Format(PRODUCT_LIST_KEY_FORMAT, id)

			Dim productIDs As IList(Of Integer) = CType(_catalogCache.Get(listKey), IList(Of Integer))
			Dim products As IList(Of Product) = Nothing

			If productIDs IsNot Nothing Then
				products = New List(Of Product)(productIDs.Count)

				For Each productID As Integer In productIDs
					products.Add(CType(_catalogCache.Get(String.Format(PRODUCT_KEY_FORMAT, productID)), Product))
				Next productID
			Else
				products = DataLayer.CatalogProvider.GetProductsByCategory(id)

				If products IsNot Nothing AndAlso products.Count > 0 Then
					productIDs = New List(Of Integer)(products.Count)
					For Each p As Product In products
						productIDs.Add(p.ProductID)

						_catalogCache.Put(String.Format(PRODUCT_KEY_FORMAT, p.ProductID), p)
					Next p

					_catalogCache.Put(listKey, productIDs)
				End If
			End If

			Return products
		End Function

		Public Function GetPopularProducts() As IList(Of Product) Implements ICatalogCache.GetPopularProducts
			Dim productIDs As IList(Of Integer) = CType(_reportCache.Get(POPULAR_LIST_KEY), IList(Of Integer))
			Dim products As IList(Of Product) = Nothing

			If productIDs IsNot Nothing Then
				products = New List(Of Product)(productIDs.Count)

				For Each productID As Integer In productIDs
					products.Add(CType(_catalogCache.Get(String.Format(PRODUCT_KEY_FORMAT, productID)), Product))
				Next productID
			Else
				products = DataLayer.CatalogProvider.GetPopularProducts()

				If products IsNot Nothing AndAlso products.Count > 0 Then
					productIDs = New List(Of Integer)(products.Count)
					For Each p As Product In products
						productIDs.Add(p.ProductID)

						_catalogCache.Put(String.Format(PRODUCT_KEY_FORMAT, p.ProductID), p)
					Next p

					_reportCache.Put(POPULAR_LIST_KEY, productIDs)
				End If
			End If

			Return products
		End Function

		Public Function GetPopularProductsByCategory(ByVal id As Integer) As IList(Of Product) Implements ICatalogCache.GetPopularProductsByCategory
			Dim listKey As String = String.Format(POPULAR_LIST_KEY_FORMAT, id)

			Dim productIDs As IList(Of Integer) = CType(_reportCache.Get(listKey), IList(Of Integer))
			Dim products As IList(Of Product) = Nothing

			If productIDs IsNot Nothing Then
				products = New List(Of Product)(productIDs.Count)

				For Each productID As Integer In productIDs
					products.Add(CType(_catalogCache.Get(String.Format(PRODUCT_KEY_FORMAT, productID)), Product))
				Next productID
			Else
				products = DataLayer.CatalogProvider.GetPopularProductsByCategory(id)

				If products IsNot Nothing AndAlso products.Count > 0 Then
					productIDs = New List(Of Integer)(products.Count)
					For Each p As Product In products
						productIDs.Add(p.ProductID)
						_catalogCache.Put(String.Format(PRODUCT_KEY_FORMAT, p.ProductID), p)
					Next p

					_reportCache.Put(listKey, productIDs)
				End If
			End If

			Return products
		End Function

		Public Function GetAlsoPurchasedProducts(ByVal id As Integer) As IList(Of Product) Implements ICatalogCache.GetAlsoPurchasedProducts
			Dim listKey As String = String.Format(ALSO_PURCHASED_KEY_FORMAT, id)

			Dim productIDs As IList(Of Integer) = CType(_reportCache.Get(listKey), IList(Of Integer))
			Dim products As IList(Of Product) = Nothing

			If productIDs IsNot Nothing Then
				products = New List(Of Product)(productIDs.Count)

				For Each productID As Integer In productIDs
					products.Add(CType(_catalogCache.Get(String.Format(PRODUCT_KEY_FORMAT, productID)), Product))
				Next productID
			Else
				products = DataLayer.CatalogProvider.GetAlsoPurchasedProducts(id)

				If products IsNot Nothing AndAlso products.Count > 0 Then
					productIDs = New List(Of Integer)(products.Count)
					For Each p As Product In products
						productIDs.Add(p.ProductID)

						_catalogCache.Put(String.Format(PRODUCT_KEY_FORMAT, p.ProductID), p)
					Next p

					_reportCache.Put(listKey, productIDs)
				End If
			End If

			Return products
		End Function
	End Class
End Namespace
